﻿using System.Collections.Generic;
using System.IO;
using UnityEngine;

namespace xgm.guru.w3_to_unity
{
    internal sealed class MdxModel
    {
        private MdxModel()
        {
            //Empty
        }

        public void Load(MdxReader reader, Model model)
        {
            var pivotPoints = new List<Vector3>();
            
            reader.ExpectTag("MDLX");
            while (true)
            {
                string tag;
                try
                {
                    if (reader.eof) break;
                    tag = reader.ReadTag();
                }
                catch (EndOfStreamException)
                {
                    break;
                }

                switch (tag)
                {
                    case "VERS": { CModelVersion.Instance.Load(reader, model); break; }
                    case "MODL": { CModelInfo.Instance.Load(reader, model); break; }
                    case "SEQS": { CSequence.Instance.LoadAll(reader, model); break; }
                    case "GLBS": { CGlobalSequence.Instance.LoadAll(reader, model); break; }
                    case "TEXS": { CTexture.Instance.LoadAll(reader, model); break; }
                    case "MTLS": { CMaterial.Instance.LoadAll(reader, model); break; }
                    case "TXAN": { CTextureAnimation.Instance.LoadAll(reader, model); break; }
                    case "GEOS": { CGeoset.Instance.LoadAll(reader, model); break; }
                    case "GEOA": { CGeosetAnimation.Instance.LoadAll(reader, model); break; }
                    case "BONE": { CBone.Instance.LoadAll(reader, model); break; }
                    case "LITE": { CLight.Instance.LoadAll(reader, model); break; }
                    case "HELP": { CHelper.Instance.LoadAll(reader, model); break; }
                    case "ATCH": { CAttachment.Instance.LoadAll(reader, model); break; }
                    case "PIVT": { CPivotPoint.Instance.Load(reader, model, pivotPoints); break; }
                    case "PREM": { CParticleEmitter.Instance.LoadAll(reader, model); break; }
                    case "PRE2": { CParticleEmitter2.Instance.LoadAll(reader, model); break; }
                    case "RIBB": { CRibbonEmitter.Instance.LoadAll(reader, model); break; }
                    case "EVTS": { CEvent.Instance.LoadAll(reader, model); break; }
                    case "CAMS": { CCamera.Instance.LoadAll(reader, model); break; }
                    case "CLID": { CCollisionShape.Instance.LoadAll(reader, model); break; }
                    case "META": { CMetaData.Instance.Load(reader, model); break; }

                    default:
                        {
                            reader.Skip(reader.ReadInt32());
                            break;
                        }
                }
            }

            SetPivotPoints(model, pivotPoints);
        }

        public void Save(CSaver Saver, MdxModel.CModel Model, string Info)
        {
            var PivotPointList = new System.Collections.Generic.List<Primitives.CVector3>();

            GetPivotPoints(Model, PivotPointList);
            Saver.WriteTag("MDLX");

            Saver.WriteTag("INFO");
            Saver.PushLocation();
            Saver.WriteString(Info, Info.Length);
            Saver.PopExclusiveLocation();

            CModelVersion.Instance.Save(Saver, Model);
            CModelInfo.Instance.Save(Saver, Model);
            CSequence.Instance.SaveAll(Saver, Model);
            CGlobalSequence.Instance.SaveAll(Saver, Model);
            CTexture.Instance.SaveAll(Saver, Model);
            CMaterial.Instance.SaveAll(Saver, Model);
            CTextureAnimation.Instance.SaveAll(Saver, Model);
            CGeoset.Instance.SaveAll(Saver, Model);
            CGeosetAnimation.Instance.SaveAll(Saver, Model);
            CBone.Instance.SaveAll(Saver, Model);
            CLight.Instance.SaveAll(Saver, Model);
            CHelper.Instance.SaveAll(Saver, Model);
            CAttachment.Instance.SaveAll(Saver, Model);
            CPivotPoint.Instance.Save(Saver, Model, PivotPointList);
            CParticleEmitter.Instance.SaveAll(Saver, Model);
            CParticleEmitter2.Instance.SaveAll(Saver, Model);
            CRibbonEmitter.Instance.SaveAll(Saver, Model);
            CEvent.Instance.SaveAll(Saver, Model);
            CCamera.Instance.SaveAll(Saver, Model);
            CCollisionShape.Instance.SaveAll(Saver, Model);
            CMetaData.Instance.Save(Saver, Model);
        }

        private void SetPivotPoints(MdxModel model, IEnumerable<Vector3> pivotPoints)
        {
            int Index = 0;
            int NrOfNodes = model.Nodes.Count;

            foreach (Primitives.CVector3 PivotPoint in pivotPoints)
            {
                if (Index >= NrOfNodes) break;
                model.Nodes[Index].PivotPoint = PivotPoint;
                Index++;
            }
        }

        private void GetPivotPoints(MdxModel model, ICollection<Vector3> pivotPoints)
        {
            foreach (MdxModel.INode Node in model.Nodes)
            {
                pivotPoints.Add(Node.PivotPoint);
            }
        }

        public static MdxModel instance
        {
            get
            {
                return CSingleton.instance;
            }
        }

        private static class CSingleton
        {
            static CSingleton() { }
            public static MdxModel instance = new MdxModel();
        }
    }
}
